PROGRAM FUND16;

(*  Climate Framework for Uncertainty, Negotiation and Distribution       *)
(*  Version 1.60  -  1 July 1997                                          *)
(*  Richard S.J. Tol                                                      *)
(*  Institute for Environmental Studies, Vrije Universiteit, Amsterdam    *)

(*  The Climate FUND is an integrated climate-economy model to calculate  *)
(*  the socio-economic impacts of both climate change and greenhouse gas  *)
(*  emission reduction, and to weight these two against one another in    *)
(*  order to determine the optimal abatement strategy in several game     *)
(*  theoretic settings. FUND is a highly parameterised, scenario-driven,  *)
(*  closed-loop model of the economies of nine major world regions and    *)
(*  their interactions with the climate. The framework includes a range   *)
(*  of climate change impacts, including intangible damages and forced    *)
(*  migration, and a suite of emission reduction measures, including      *)
(*  interregional capital transfers. The development of the Climate FUND  *)
(*  is sponsored by the Dutch National Research Programme: Global Air     *)
(*  Pollution and Climate Change, phase I.                                *)

USES
    Crt, RP160, GenRand;

CONST
     scratch = false;                  (* start with no measures ? *)
     CBA = false;                     (* cost-benefit analysis ? *)
     CEA = false;                      (* cost-effectiveness analysis ? *)
     captrans = false;                (* capital transfers ? *)
     cooperate = false;               (* cooperate ? *)
     uncertain = false;               (* uncertainty ? *)
     sequent = true;                 (* sequential ? *)
     scengmt = false;                 (* scenario global mean temperature ? *)

     consleak = 0.25;                 (* consumption leak *)
     compth = 100;                    (* compensation threshold *)

     noper = 20;

     ampl = 4;
     altruism = 0.00;
     runs = 1;

     bregst = 0.0001;
     bforst = 1.00;
     btaxst = 0.0001;

     CBaU = 1792.3352;
     CMax = 550;

TYPE
    measure = regarr;
    matmeas = ARRAY [1..NoReg,1..NOI] OF double;
    boolmea = ARRAY [1..NoReg,1..NOI] OF boolean;

VAR
   boolmeas: boolmea;
   regstep, forstep, taxstep: double;
   c2000,c2025,c2050,c2075,c2100,c2200: double;
   rhos, rhoa, rhohc, rhoh, rhowt, rhowi, rhoe, rhoi, rhocp, rhodl: double;
   optper: word;
   globviol: boolean;
   emmrec, count: regarr;

FUNCTION Log(x: double): double;
(* returns ln(x) *)

BEGIN
     IF (x > 1.0E-308) AND (x < 1.0E308) THEN
        log:= ln(x)
     ELSE IF x < 1.0E-308 THEN
          log:= ln(1.0E-308)
     ELSE
         log:= ln(1.0E38)
END;


FUNCTION Expo(x: double): double;
(* returns exp(x) *)

BEGIN
     IF (x < 709) AND (x > -11356) THEN
        expo:= exp(x)
     ELSE IF x < -11356 THEN
          expo:= exp(-11356)
     ELSE
         expo:= exp(709)
END;

FUNCTION Root(x: double): double;
(* returns sqrt(x) *)

BEGIN
     IF x < 1.0E308 THEN
        root:= sqrt(x)
     ELSE
         root:= 1.0E154
END;

FUNCTION Quad(x: double): double;
(* returns x^2 *)

BEGIN
     IF x < 1.0E154 THEN
        quad:= sqr(x)
     ELSE
         quad:= 1.0E308
END;

FUNCTION Pow(x,y: double): double;
(* returns x^y *)

BEGIN
     pow:= expo(y*log(x))
END;

FUNCTION Interact(M,N: double): double;
(* interaction radiative forcing methane and nitrous oxide (in ppbv) *)

VAR hulp: double;

BEGIN
     hulp:= 1 + pow(M*N,0.75)*2.01E-5 + pow(M*N,1.52)*M*5.31E-15;
     interact:= 0.47*log(hulp)
END;

PROCEDURE InitRho;

BEGIN
     rhos:=  1-1/100;
     rhoa:=  1-1/10;
     rhohc:= 1-1/15;
     rhoh:=  1-1/5;
     rhoi:=  1-1/5;
     rhoe:=  1-1/5;
     rhowt:= 1-1/10;
     rhowi:= 1-1/50;
     rhodl:= 1-1/50;
     rhocp:= 1-1/50
END;

PROCEDURE InitRegMult;

VAR i: word;

BEGIN
     economy[1].regmp:= 4.4788/2.5;
     economy[2].regmp:= 4.1401/2.5;
     economy[3].regmp:= 3.0492/2.5;
     economy[4].regmp:= 4.4397/2.5;
     economy[5].regmp:= 4.6320/2.5;
     economy[6].regmp:= 3.5218/2.5;
     economy[7].regmp:= 3.4416/2.5;
     economy[8].regmp:= 3.9997/2.5;
     economy[9].regmp:= 3.9804/2.5;

     FOR i:= 1 TO NoReg DO
         economy[i].regmp:= 1;

     FOR i:= 1 TO NoReg DO
         economy[i].regsq:= economy[i].regmp*economy[i].regmp
END;

PROCEDURE Initialise;

VAR i:word;

BEGIN
     InitRho;
     InitRegMult;
     optper:= 0;
     C2200:= CBaU;
     FOR i:= 1 TO NoReg DO
         WITH Economy[i] DO BEGIN
              new(ypcgrowth);
              new(pgrowth);
              new(urbanp);
              new(aeei);
              new(acei);
              new(forestemm)
         END;
     WITH Climate DO BEGIN
          decay[1]:= 1;
          decay[2]:= exp(-1/lifeco2);
          decay[3]:= exp(-1/lifeco3);
          decay[4]:= exp(-1/lifeco4);
          decay[5]:= exp(-1/lifeco5);
          frac[1]:= 0.13;
          frac[2]:= 0.20;
          frac[3]:= 0.32;
          frac[4]:= 0.25;
          frac[5]:= 0.10;
          ch4decay:= ch4dec;
          n2odecay:= n2odec;
          new(addco2);
          new(addch4);
          new(addn2o);
          new(scentemp);
          new(ch4em);
          new(n2oem);
          new(acco2);
          new(acch4);
          new(acn2o)
     END;

     Economy[1].name:= 'OECD-A';
     Economy[2].name:= 'OECD-E';
     Economy[3].name:= 'OECD-P';
     Economy[4].name:= 'CEE&SU';
     Economy[5].name:= 'M-E';
     Economy[6].name:= 'L-A';
     Economy[7].name:= 'S&SEA';
     Economy[8].name:= 'CPA';
     Economy[9].name:= 'AFR';

     Economy[1].captrans:= 0;
     Economy[2].captrans:= 0;
     Economy[3].captrans:= 0;
     Economy[4].captrans:= 0;
     Economy[5].captrans:= 0;
     Economy[6].captrans:= 0;
     Economy[7].captrans:= 0;
     Economy[8].captrans:= 0;
     Economy[9].captrans:= 0;

     Economy[1].compens:= 0;
     Economy[2].compens:= 0;
     Economy[3].compens:= 0;
     Economy[4].compens:= 0;
     Economy[5].compens:= 0;
     Economy[6].compens:= 0;
     Economy[7].compens:= 0;
     Economy[8].compens:= 0;
     Economy[9].compens:= 0;

     Economy[1].ypccorr:= 0.421855;
     Economy[2].ypccorr:= 0.263729;
     Economy[3].ypccorr:= 0.244423;
     Economy[4].ypccorr:= 0.306673;
     Economy[5].ypccorr:= 0.373133;
     Economy[6].ypccorr:= 0.437524;
     Economy[7].ypccorr:= 0.196623;
     Economy[8].ypccorr:= 0.121762;
     Economy[9].ypccorr:= 0.724458;

     ReadIncome;
     ReadPopulation;
     ReadYpCGrowth;
     ReadAgrish;
     ReadPGrowth;
     ReadUrbanPop;
     ReadEnergint;
     ReadACEI;
     ReadForestEmm;
     ReadForParam;
     ReadDiscRate;
     ReadTaxPar;
     ReadHurrLife;
     ReadHurrProp;
     ReadFloodDam;
     ReadWindDam;
     ReadHeat;
     ReadCold;
     ReadMalaria;
     ReadWetland;
     ReadDryland;
     ReadProtect;
     ReadSpecies;
     ReadAgricult;
     ReadLeave;
     ReadEnter;
     ReadAddCO2;
     ReadAddCH4;
     ReadAddN2O;
     ReadCH4Em;
     ReadN2OEm;
     ReadScenGMT;
     ReadNCEmmRestr;
     ReadCEmmRestr;
     WITH climate DO BEGIN
          temps:= modet/6.3/ln(2);
          hurrs:= modeh;
          seas:= modes;
          delaysea:= 1/lifesea;
          delaytemp:= 1/lifetemp;
          floods:= modef;
          winds:= modew
     END;
     bstsq:= quad(baset);
     dbstsq:= quad(dbst);
     bsssq:= quad(bases);
     dbsssq:= quad(dbss);
     regstep:= bregst;
     forstep:= bforst;
     taxstep:= btaxst;
     FOR i:= 1 TO NoReg DO
         WITH economy[i] DO BEGIN
              ecgrmp:= 1;
              popgrmp:= 1;
              aeeimp:= 1;
              aceimp:= 1;
              foremp:= 1;
              taxmp:= 1;
              formp:= 1;
              specmp:= 1;
              heatmp:= 1;
              coldmp:= 1;
              malarmp:= 1;
              agrimp:= 1;
              hurrlmp:= 1;
              hurrpmp:= 1;
              floodmp:= 1;
              windmp:= 1;
              migmp:= 1;
              protmp:= 1;
              drymp:= 1;
              wetmp:= 1;
              voslmp:= 1
         END
END;

PROCEDURE ReInit;

VAR i: word;

BEGIN
     WITH economy[1] DO   (* emf *)
          emissint:= (*0.2567013*) 0.283/0.5529;
     WITH economy[2] DO
          emissint:= (*0.139178*)  0.220/0.4699;
     WITH economy[3] DO
          emissint:= (*0.1144091*) 0.118/0.9341;
     WITH economy[4] DO
          emissint:= (*0.4464261*) 0.413/0.8092;
     WITH economy[5] DO
          emissint:= (*0.4752512*) 0.455/6.0675;
     WITH economy[6] DO
          emissint:= (*0.3097415*) 0.349/0.9486;
     WITH economy[7] DO
          emissint:= (*0.4460469*) 0.492/1.3236;
     WITH economy[8] DO
          emissint:= (*1.6008936*) 1.732/1.6073;
     WITH economy[9] DO
          emissint:= (*0.4555767*) 0.440/1.6773;
     totalpop:= 0;
     FOR i:= 1 TO NoReg DO
         WITH economy[i] DO BEGIN
              inc0:= incbase + captrans;
              income:= inc0;
              population:= pop0;
              welf0:= log(1000*(1-consleak)*inc0/pop0);
              welfare:= welf0;
              YpC:= 1000*inc0/pop0;                     (* ????????? *)
              YpC0:= YpC;
              energint:= 1;                             (* confirmed *)
              energuse:= inc0;
              ryg:= 0;
              reei:= 0;                                 (* confirmed *)
              rcei:= 0;                                 (* confirmed *)
              emission:= emissint*energuse;
              emiss0:= emission;                        (* confirmed *)
              totalpop:= totalpop + pop0;
              cdr:= 1;
              y:= 1;
              v:= 1;
              p:= 1;
              pg:= popgrmp*pgrowth^[0]/100;
              spec:= 0;
              deadht:= 0;               (*?*)
              deadhr:= 0;
              deadct:= 0;
              deadcr:= 0;
              stress:= 0;
              malaria:= 0;
              malcost:= 0;
              agri1:= 0;
              agri2:= 0;
              hurr1:= 0;
              hurr2:= 0;
              hurr3:= 0;
              hurr4:= 0;
              hurr5:= 0;
              flooddam:= 0;
              winddam:= 0;
              enter:= 0;
              leav:= 0;
              enterc:= 0;
              leavec:= 0;
              coastm:= 0;
              dryld:= 0;
              coastc:= 0;
              drylt:= 0;
              welti:= 0;
              weltt:= 0;
              eloss:= 0;
              sloss:= 0
     END;
     WITH climate DO BEGIN
          temp:= 0.20;
          scentemp^[0]:= temp;
          tsq:= 0;
          dt:= 0;
          dtsq:= 0;
          sea:= 0;
          ssq:= 0;
          ds:= 0;
          dssq:= 0;
          hurr:= 0;
          wind:= 0;
          flood:= 0;
          cbox[1]:= 283.53;
          cbox[2]:= 5.62;
          cbox[3]:= 6.29;
          cbox[4]:= 2.19;
          cbox[5]:= 0.15;
          acco2^[0]:= 0;
          acch4^[0]:= 1222;
          acn2o^[0]:= 296;
          FOR i:= 1 TO 5 DO
              acco2^[0]:= acco2^[0] + cbox[i];
          ch4n2o:= interact(ch4pre,n2opre)
     END
END;

(* TAX PROCEDURES *)

PROCEDURE SetTax1(i,j: word);
(* sets tax country I in decade J to zero *)

BEGIN
     economy[i].tax[j]:= 0;
END;

PROCEDURE SetTax2;
(* sets all taxes to zero *)

VAR i, j: word;

BEGIN
     FOR i:= 1 TO NoReg DO
         FOR j:= 0 TO CtrMax DO
             economy[i].tax[j]:= 0;
END;

PROCEDURE SetTax3(i: word; T: double);

VAR j: word;

BEGIN
     IF NOT sequent THEN
        FOR j:= 0 TO CtrMax DO
            economy[i].tax[j]:= T
     ELSE
         economy[i].tax[optper]:= T;
END;

PROCEDURE SetTax4(T: measure);

VAR i,j: word;

BEGIN
     IF NOT sequent THEN
        FOR i:= 1 TO NoReg DO
            FOR j:= 0 TO CtrMax DO
                economy[i].tax[j]:= T[i]
     ELSE
         FOR i:= 1 TO NoReg DO
             economy[i].tax[optper]:= T[i]
END;

PROCEDURE SetTax5(i,j: word; T: double);

BEGIN
     economy[i].tax[j]:= T
END;

PROCEDURE ObjTax(i: word);
(* changes taxes and influences for economy I *)

VAR j: word;

BEGIN
     IF NOT sequent THEN
        FOR j:= 0 TO CtrMax DO
            economy[i].tax[j]:= economy[i].tax[j] + taxstep
     ELSE
         economy[i].tax[optper]:= economy[i].tax[optper] + taxstep
END;

PROCEDURE TaxObj(i: word);
(* changes taxes and influences one taxstep back for economy I *)

VAR j: word;

BEGIN
     IF NOT sequent THEN
        FOR j:= 0 TO CtrMax DO
            economy[i].tax[j]:= economy[i].tax[j] - taxstep
     ELSE
         economy[i].tax[optper]:= economy[i].tax[optper] - taxstep
END;

(* FORESTRY PROCEDURES *)

PROCEDURE SetInflFor(VAR country: regrec; t: word);
(* determines the influence of COUNTRY.FORESTRY on the economy of COUNTRY *)

VAR fca, fcd, fa, fd, th1, th2: double;

BEGIN
     WITH country DO BEGIN
          IF forestry[optper] > 0 THEN BEGIN
             th1:= forconst - forlinear + root(quad(forconst-forlinear)-4*forconst*forsquare);
             th1:= th1/2/forsquare;
             th2:= (3*forconst/2 - forlinear)/2/forsquare;
             IF forestry[optper] < (th1 + th2) THEN BEGIN
                fd:= forestry[optper];
                fa:= 0
             END
             ELSE IF forestry[optper] < (2*th1 + th2) THEN BEGIN
                  fd:= th1 + th2;
                  fa:= forestry[optper] - fd
             END
             ELSE BEGIN
                  fa:= th1;
                  fd:= forestry[optper] - fa
             END;
             IF fd > foremp*forestemm^[t] THEN BEGIN    (* !!! *)
                fd:= foremp*forestemm^[t];
                fa:= forestry[optper] - fd
             END;
             IF fd > 0 THEN BEGIN
                fcd:= 2*(forconst + forlinear*fd + forsquare*quad(fd))/3;
                IF fcd < forconst*fd THEN
                   fcd:= 2*forconst*fd/3
             END
             ELSE
                 fcd:= 0;
             IF fa > 0 THEN BEGIN
                fca:= forconst + forlinear*fa + forsquare*quad(fa);
                IF fca < forconst*fa THEN
                   fca:= forconst*fa
             END
             ELSE
                 fca:= 0;
             forcost:= fca + fcd
          END
          ELSE
              forcost:= 0
     END
END;

PROCEDURE SetFor1(i,j: word);
(* sets forestry in country I in decade J to zero *)

BEGIN
     economy[i].forestry[j]:= 0;
END;

PROCEDURE SetFor2;
(* sets all forestry to zero *)

VAR i, j: word;

BEGIN
     FOR i:= 1 TO NoReg DO
         FOR j:= 1 TO CtrMax DO
             economy[i].forestry[j]:= 0;
END;

PROCEDURE SetFor3(i: word; T: double);

VAR j: word;

BEGIN
     IF NOT sequent THEN
        FOR j:= 0 TO CtrMax DO
            economy[i].forestry[j]:= T
     ELSE
         economy[i].forestry[optper]:= T;
END;

PROCEDURE SetFor4(T: measure);

VAR i,j: word;

BEGIN
     IF NOT sequent THEN
        FOR i:= 1 TO NoReg DO
            FOR j:= 0 TO CtrMax DO
                economy[i].forestry[j]:= T[i]
     ELSE
         FOR i:= 1 TO NoReg DO
             economy[i].forestry[optper]:= T[i]
END;

PROCEDURE ObjFor(i: word);
(* changes forestry in economy I *)

VAR j: word;

BEGIN
     IF NOT sequent THEN
        FOR j:= 0 TO CtrMax DO
            economy[i].forestry[j]:= economy[i].forestry[j] + forstep
     ELSE
         economy[i].forestry[optper]:= economy[i].forestry[optper] + forstep
END;

PROCEDURE ForObj(i: word);
(* changes forestry one forstep back for economy I *)

VAR j: word;

BEGIN
     IF NOT sequent THEN
        FOR j:= 0 TO CtrMax DO
            economy[i].forestry[j]:= economy[i].forestry[j] - forstep
     ELSE
         economy[i].forestry[optper]:= economy[i].forestry[optper] - forstep
END;

(* SETTINGS *)

PROCEDURE SetMeasures1;
(* sets all measures to zero *)

BEGIN
     SetTax2;
     SetFor2
END;

PROCEDURE SetMeasures2;
(* sets the regional abatement measures according to an exogenously specified
   prescript *)

VAR i: word;

BEGIN
     WITH Economy[1] DO BEGIN
          tax[0]:=  0.0005;
          tax[1]:=  0.0005;
          tax[2]:=  0.0004;
          tax[3]:=  0.0004;
          tax[4]:=  0.0004;
          tax[5]:=  0.0003;
          tax[6]:=  0.0003;
          tax[7]:=  0.00;
          tax[8]:=  0.00;
          tax[9]:=  0.00;
          tax[10]:= 0.00;
          tax[11]:= 0.00;
          tax[12]:= 0.00;
          tax[13]:= 0.00;
          tax[14]:= 0.00;
          tax[15]:= 0.00;
          tax[16]:= 0.00;
          tax[17]:= 0.00;
          tax[18]:= 0.00;
          tax[19]:= 0.00;
          tax[20]:= 0.00;

          forestry[0]:= 0.0;
          forestry[1]:= 0.0;
          forestry[2]:= 0.0;
          forestry[3]:= 0.0;
          forestry[4]:= 0.0;
          forestry[5]:= 0.0;
          forestry[6]:= 0.0;
          forestry[7]:= 0.0;
          forestry[8]:= 0.0;
          forestry[9]:= 0.0;
          forestry[10]:= 0.0;
          forestry[11]:= 0.0;
          forestry[12]:= 0.0;
          forestry[13]:= 0.0;
          forestry[14]:= 0.0;
          forestry[15]:= 0.0;
          forestry[16]:= 0.0;
          forestry[17]:= 0.0;
          forestry[18]:= 0.0;
          forestry[19]:= 0.0;
          forestry[20]:= 0.0
     END;
     WITH Economy[2] DO BEGIN
          tax[0]:=  0.0005;
          tax[1]:=  0.0005;
          tax[2]:=  0.0004;
          tax[3]:=  0.0004;
          tax[4]:=  0.0003;
          tax[5]:=  0.0003;
          tax[6]:=  0.0002;
          tax[7]:=  0.00;
          tax[8]:=  0.00;
          tax[9]:=  0.00;
          tax[10]:= 0.00;
          tax[11]:= 0.00;
          tax[12]:= 0.00;
          tax[13]:= 0.00;
          tax[14]:= 0.00;
          tax[15]:= 0.00;
          tax[16]:= 0.00;
          tax[17]:= 0.00;
          tax[18]:= 0.00;
          tax[19]:= 0.00;
          tax[20]:= 0.00;

          forestry[0]:= 0.0;
          forestry[1]:= 0.0;
          forestry[2]:= 0.0;
          forestry[3]:= 0.0;
          forestry[4]:= 0.0;
          forestry[5]:= 0.0;
          forestry[6]:= 0.0;
          forestry[7]:= 0.0;
          forestry[8]:= 0.0;
          forestry[9]:= 0.0;
          forestry[10]:= 0.0;
          forestry[11]:= 0.0;
          forestry[12]:= 0.0;
          forestry[13]:= 0.0;
          forestry[14]:= 0.0;
          forestry[15]:= 0.0;
          forestry[16]:= 0.0;
          forestry[17]:= 0.0;
          forestry[18]:= 0.0;
          forestry[19]:= 0.0;
          forestry[20]:= 0.0
     END;
     WITH Economy[3] DO BEGIN
          tax[0]:=  0.0001;
          tax[1]:=  0.0001;
          tax[2]:=  0.0001;
          tax[3]:=  0.0001;
          tax[4]:=  0.0001;
          tax[5]:=  0.0001;
          tax[6]:=  0.0001;
          tax[7]:=  0.00;
          tax[8]:=  0.00;
          tax[9]:=  0.00;
          tax[10]:= 0.00;
          tax[11]:= 0.00;
          tax[12]:= 0.00;
          tax[13]:= 0.00;
          tax[14]:= 0.00;
          tax[15]:= 0.00;
          tax[16]:= 0.00;
          tax[17]:= 0.00;
          tax[18]:= 0.00;
          tax[19]:= 0.00;
          tax[20]:= 0.00;

          forestry[0]:= 0.0;
          forestry[1]:= 0.0;
          forestry[2]:= 0.0;
          forestry[3]:= 0.0;
          forestry[4]:= 0.0;
          forestry[5]:= 0.0;
          forestry[6]:= 0.0;
          forestry[7]:= 0.0;
          forestry[8]:= 0.0;
          forestry[9]:= 0.0;
          forestry[10]:= 0.0;
          forestry[11]:= 0.0;
          forestry[12]:= 0.0;
          forestry[13]:= 0.0;
          forestry[14]:= 0.0;
          forestry[15]:= 0.0;
          forestry[16]:= 0.0;
          forestry[17]:= 0.0;
          forestry[18]:= 0.0;
          forestry[19]:= 0.0;
          forestry[20]:= 0.0
     END;
     WITH Economy[4] DO BEGIN
          tax[0]:=  0.00;
          tax[1]:=  0.00;
          tax[2]:=  0.00;
          tax[3]:=  0.00;
          tax[4]:=  0.00;
          tax[5]:=  0.00;
          tax[6]:=  0.00;
          tax[7]:=  0.00;
          tax[8]:=  0.00;
          tax[9]:=  0.00;
          tax[10]:= 0.00;
          tax[11]:= 0.00;
          tax[12]:= 0.00;
          tax[13]:= 0.00;
          tax[14]:= 0.00;
          tax[15]:= 0.00;
          tax[16]:= 0.00;
          tax[17]:= 0.00;
          tax[18]:= 0.00;
          tax[19]:= 0.00;
          tax[20]:= 0.00;

          forestry[0]:= 0.0;
          forestry[1]:= 0.0;
          forestry[2]:= 0.0;
          forestry[3]:= 0.0;
          forestry[4]:= 0.0;
          forestry[5]:= 0.0;
          forestry[6]:= 0.0;
          forestry[7]:= 0.0;
          forestry[8]:= 0.0;
          forestry[9]:= 0.0;
          forestry[10]:= 0.0;
          forestry[11]:= 0.0;
          forestry[12]:= 0.0;
          forestry[13]:= 0.0;
          forestry[14]:= 0.0;
          forestry[15]:= 0.0;
          forestry[16]:= 0.0;
          forestry[17]:= 0.0;
          forestry[18]:= 0.0;
          forestry[19]:= 0.0;
          forestry[20]:= 0.0
     END;
     WITH Economy[5] DO BEGIN
          tax[0]:=  0.0003;
          tax[1]:=  0.0003;
          tax[2]:=  0.0002;
          tax[3]:=  0.0002;
          tax[4]:=  0.0002;
          tax[5]:=  0.0002;
          tax[6]:=  0.0002;
          tax[7]:=  0.00;
          tax[8]:=  0.00;
          tax[9]:=  0.00;
          tax[10]:= 0.00;
          tax[11]:= 0.00;
          tax[12]:= 0.00;
          tax[13]:= 0.00;
          tax[14]:= 0.00;
          tax[15]:= 0.00;
          tax[16]:= 0.00;
          tax[17]:= 0.00;
          tax[18]:= 0.00;
          tax[19]:= 0.00;
          tax[20]:= 0.00;

          forestry[0]:= 0.0;
          forestry[1]:= 0.0;
          forestry[2]:= 0.0;
          forestry[3]:= 0.0;
          forestry[4]:= 0.0;
          forestry[5]:= 0.0;
          forestry[6]:= 0.0;
          forestry[7]:= 0.0;
          forestry[8]:= 0.0;
          forestry[9]:= 0.0;
          forestry[10]:= 0.0;
          forestry[11]:= 0.0;
          forestry[12]:= 0.0;
          forestry[13]:= 0.0;
          forestry[14]:= 0.0;
          forestry[15]:= 0.0;
          forestry[16]:= 0.0;
          forestry[17]:= 0.0;
          forestry[18]:= 0.0;
          forestry[19]:= 0.0;
          forestry[20]:= 0.0
     END;
     WITH Economy[6] DO BEGIN
          tax[0]:=  0.0006;
          tax[1]:=  0.0006;
          tax[2]:=  0.0005;
          tax[3]:=  0.0005;
          tax[4]:=  0.0005;
          tax[5]:=  0.0005;
          tax[6]:=  0.0005;
          tax[7]:=  0.00;
          tax[8]:=  0.00;
          tax[9]:=  0.00;
          tax[10]:= 0.00;
          tax[11]:= 0.00;
          tax[12]:= 0.00;
          tax[13]:= 0.00;
          tax[14]:= 0.00;
          tax[15]:= 0.00;
          tax[16]:= 0.00;
          tax[17]:= 0.00;
          tax[18]:= 0.00;
          tax[19]:= 0.00;
          tax[20]:= 0.00;

          forestry[0]:= 0.0;
          forestry[1]:= 0.0;
          forestry[2]:= 0.0;
          forestry[3]:= 0.0;
          forestry[4]:= 0.0;
          forestry[5]:= 0.0;
          forestry[6]:= 0.0;
          forestry[7]:= 0.0;
          forestry[8]:= 0.0;
          forestry[9]:= 0.0;
          forestry[10]:= 0.0;
          forestry[11]:= 0.0;
          forestry[12]:= 0.0;
          forestry[13]:= 0.0;
          forestry[14]:= 0.0;
          forestry[15]:= 0.0;
          forestry[16]:= 0.0;
          forestry[17]:= 0.0;
          forestry[18]:= 0.0;
          forestry[19]:= 0.0;
          forestry[20]:= 0.0
     END;
     WITH Economy[7] DO BEGIN
          tax[0]:=  0.0015;
          tax[1]:=  0.0015;
          tax[2]:=  0.0014;
          tax[3]:=  0.0014;
          tax[4]:=  0.0013;
          tax[5]:=  0.0013;
          tax[6]:=  0.0012;
          tax[7]:=  0.00;
          tax[8]:=  0.00;
          tax[9]:=  0.00;
          tax[10]:= 0.00;
          tax[11]:= 0.00;
          tax[12]:= 0.00;
          tax[13]:= 0.00;
          tax[14]:= 0.00;
          tax[15]:= 0.00;
          tax[16]:= 0.00;
          tax[17]:= 0.00;
          tax[18]:= 0.00;
          tax[19]:= 0.00;
          tax[20]:= 0.00;

          forestry[0]:= 0.0;
          forestry[1]:= 0.0;
          forestry[2]:= 0.0;
          forestry[3]:= 0.0;
          forestry[4]:= 0.0;
          forestry[5]:= 0.0;
          forestry[6]:= 0.0;
          forestry[7]:= 0.0;
          forestry[8]:= 0.0;
          forestry[9]:= 0.0;
          forestry[10]:= 0.0;
          forestry[11]:= 0.0;
          forestry[12]:= 0.0;
          forestry[13]:= 0.0;
          forestry[14]:= 0.0;
          forestry[15]:= 0.0;
          forestry[16]:= 0.0;
          forestry[17]:= 0.0;
          forestry[18]:= 0.0;
          forestry[19]:= 0.0;
          forestry[20]:= 0.0
     END;
     WITH Economy[8] DO BEGIN
          tax[0]:=  0.00;
          tax[1]:=  0.00;
          tax[2]:=  0.00;
          tax[3]:=  0.00;
          tax[4]:=  0.00;
          tax[5]:=  0.00;
          tax[6]:=  0.00;
          tax[7]:=  0.00;
          tax[8]:=  0.00;
          tax[9]:=  0.00;
          tax[10]:= 0.00;
          tax[11]:= 0.00;
          tax[12]:= 0.00;
          tax[13]:= 0.00;
          tax[14]:= 0.00;
          tax[15]:= 0.00;
          tax[16]:= 0.00;
          tax[17]:= 0.00;
          tax[18]:= 0.00;
          tax[19]:= 0.00;
          tax[20]:= 0.00;

          forestry[0]:= 0.0;
          forestry[1]:= 0.0;
          forestry[2]:= 0.0;
          forestry[3]:= 0.0;
          forestry[4]:= 0.0;
          forestry[5]:= 0.0;
          forestry[6]:= 0.0;
          forestry[7]:= 0.0;
          forestry[8]:= 0.0;
          forestry[9]:= 0.0;
          forestry[10]:= 0.0;
          forestry[11]:= 0.0;
          forestry[12]:= 0.0;
          forestry[13]:= 0.0;
          forestry[14]:= 0.0;
          forestry[15]:= 0.0;
          forestry[16]:= 0.0;
          forestry[17]:= 0.0;
          forestry[18]:= 0.0;
          forestry[19]:= 0.0;
          forestry[20]:= 0.0
     END;
     WITH Economy[9] DO BEGIN
          tax[0]:=  0.0005;
          tax[1]:=  0.0005;
          tax[2]:=  0.0005;
          tax[3]:=  0.0004;
          tax[4]:=  0.0004;
          tax[5]:=  0.0004;
          tax[6]:=  0.0004;
          tax[7]:=  0.00;
          tax[8]:=  0.00;
          tax[9]:=  0.00;
          tax[10]:= 0.00;
          tax[11]:= 0.00;
          tax[12]:= 0.00;
          tax[13]:= 0.00;
          tax[14]:= 0.00;
          tax[15]:= 0.00;
          tax[16]:= 0.00;
          tax[17]:= 0.00;
          tax[18]:= 0.00;
          tax[19]:= 0.00;
          tax[20]:= 0.00;

          forestry[0]:= 0.0;
          forestry[1]:= 0.0;
          forestry[2]:= 0.0;
          forestry[3]:= 0.0;
          forestry[4]:= 0.0;
          forestry[5]:= 0.0;
          forestry[6]:= 0.0;
          forestry[7]:= 0.0;
          forestry[8]:= 0.0;
          forestry[9]:= 0.0;
          forestry[10]:= 0.0;
          forestry[11]:= 0.0;
          forestry[12]:= 0.0;
          forestry[13]:= 0.0;
          forestry[14]:= 0.0;
          forestry[15]:= 0.0;
          forestry[16]:= 0.0;
          forestry[17]:= 0.0;
          forestry[18]:= 0.0;
          forestry[19]:= 0.0;
          forestry[20]:= 0.0
     END;
END;

PROCEDURE SetMeasures3(w: word);

VAR s: string;
    inp: text;
    i,j: word;

BEGIN
     str(w,s);
     assign(inp,sessionpathname+'seq'+s+'.prn');
     reset(inp);
     FOR i:= 0 TO NoPer DO BEGIN
         FOR j:= 1 TO NoReg DO
             read(inp,economy[j].tax[i]);
         readln(inp);
         FOR j:= 1 TO NoReg DO
             read(inp,economy[j].forestry[i]);
         readln(inp)
     END;
     close(inp)
END;

PROCEDURE SetMeasures4;

VAR inp: text;
    i,j: word;

BEGIN
     assign(inp,sessionpathname+'cea.prn');
     reset(inp);
     FOR i:= 0 TO NoPer DO BEGIN
         FOR j:= 1 TO NoReg DO
             read(inp,economy[j].tax[i]);
         readln(inp);
(*         FOR j:= 1 TO NoReg DO
             read(inp,economy[j].forestry[i]);
         readln(inp)*)
         FOR j:= 1 TO NoReg DO
             economy[j].forestry[i]:= 0
     END;
     close(inp)
END;

FUNCTION CheckStop(w: word): boolean;

VAR s: string;
    inp0,inp1: text;
    i,j: word;
    h0,h1: double;
    stop: boolean;

BEGIN
     str(w-1,s);
     assign(inp0,sessionpathname+'seq'+s+'.prn');
     reset(inp0);
     str(w,s);
     assign(inp1,sessionpathname+'seq'+s+'.prn');
     reset(inp1);
     stop:= true;
     FOR i:= 0 TO NoPer DO BEGIN
         FOR j:= 1 TO NoReg DO BEGIN
             read(inp0,h0);
             read(inp1,h1);
             stop:= stop AND (ABS(h1-h0) < 0.00005)
         END;
         readln(inp0);
         readln(inp1);
         FOR j:= 1 TO NoReg DO BEGIN
             read(inp0,h0);
             read(inp1,h1);
             stop:= stop AND (ABS(h1-h0) < 0.00005)
         END;
         readln(inp0);
         readln(inp1)
     END;
     close(inp0);
     close(inp1);
     checkstop:= stop
END;

(* CALCULATING PROCEDURES *)

PROCEDURE Calculate(VAR country: regrec; t: word);

VAR i, j, decade: word;
    s, h, henter, excessloss, popold, ic, dc, x: double;

BEGIN
     WITH country, climate DO BEGIN
          income:= ((1+pg)*(1+ecgrmp*ypcgrowth^[t-1]/100)-taxmp*ryg)*income;
          y:= income/inc0;
          w1950:= welfare/welf0;
          popold:= population;
          population:= (1+0.01*popgrmp*pgrowth^[t-1])*(population
                       + deadht + deadhr + deadct + deadcr);
          p:= population/pop0;
          YpC:= 1000*income/population;
          v:= ypc/ypc0;
          agrish:= agrish0*pow(v,-0.31);
          energint:= (1 - 0.01*aeeimp*aeei^[t] - reei)*energint;
          energuse:= energint*income;
          emissint:= (1 - 0.01*aceimp*acei^[t] - rcei)*emissint;
          emission:= emissint*energuse;

          IF (t mod 10 = 0) AND (t >= 40) THEN BEGIN
             decade:= trunc((t-40)/10);
             ryg:= taxpar1*tax[decade] + taxpar2*quad(tax[decade]);
             reei:= 0.5*tax[decade];
             rcei:= 0.5*tax[decade]
          END;
          setinflfor(country,t);

          ic:= (1+ypc0/ypccorr/20000)*ypc*ypccorr/ypc0/(1+ypc/20000);
          dc:= (1+ypc0/ypccorr/500)/(1+ypc/500);

          spec:= 0.5*specmp*specl*income*(regmp*abs(dt/dbst)+regsq*sqr(dt/dbst))/60*ic
                 + rhos*spec;
          deadht:= heatmp*heatt*urbanp^[t]*population*regmp*temp/baset;
          deadhr:= 0.5*heatmp*heatr*urbanp^[t]*population*(regmp*dt/dbst+regsq*sqr(dt/dbst))/60
                   + rhohc*deadhr;
          deadct:= coldmp*coldt*population*regmp*temp/baset;
          deadcr:= 0.5*coldmp*coldr*population*(regmp*dt/dbst+regsq*sqr(dt/dbst))/60
                   + rhohc*deadcr;
          dead:= deadhr+deadht-deadcr-deadct;
          stress:= voslmp*(vhlc + vhll*ypc)*dead/1000;
          malaria:= malarmp*malar*population*regmp*temp/baset*dc;
          dead:= dead + malaria;
          malcost:= voslmp*(vhlc + vhll*ypc)*malaria/1000;
          agri1:= agrimp*dagri*0.5*agrish*income*(regmp*abs(dt/dbst)+regsq*sqr(dt/dbst))/60
                  + rhoa*agri1;
          agri2:= agrimp*agri*agrish*income*regmp*temp/baset;
          x:= 1/6;
          hurr1:= x*hurrlmp*hurrl*population*regmp*hurr/baseh;
          hurr2:= x*hurrpmp*hurrp*income*regmp*hurr/baseh;
          IF dh > 0 THEN BEGIN
             hurr3:= (1-x)*0.5*hurrlmp*hurrl*population*(regmp*dh/dbsh + regsq*sqr(dh/dbsh))/60
                     + rhoh*hurr3;
             hurr4:= (1-x)*0.5*hurrpmp*hurrp*income*(regmp*dh/dbsh + regsq*sqr(dh/dbsh))/60
                     + rhoh*hurr4
          END
          ELSE BEGIN
               hurr3:= 0.2*(1-x)*0.5*hurrlmp*hurrl*population*(regmp*dh/dbsh + regsq*sqr(dh/dbsh))/60
                       + rhoh*hurr3;
               hurr4:= 0.2*(1-x)*0.5*hurrpmp*hurrp*income*(regmp*dh/dbsh + regsq*sqr(dh/dbsh))/60
                       + rhoh*hurr4
          END;
          hurr5:= voslmp*(vhlc + vhll*ypc)*(hurr1+hurr3)/1000;
          dead:= dead + hurr1 + hurr3;
          flooddam:= floodmp*floodd*income*regmp*flood/basef;
          winddam:= windmp*windd*income*regsq*sqr(wind/basew);
          henter:= 0;
          FOR j:= 1 TO 9 DO
              henter:= henter + migmp*enter0[j]*economy[j].population*(1+economy[j].ypc0/500)/(1+economy[j].ypc/500);
          enter:= henter*abs(ds/dbss)/60;
          leav:= migmp*leave*population*abs(ds/dbss)/60*dc;
          enterc:= enter*0.40*ypc + rhoi*enterc;
          leavec:= leav*3*ypc + rhoe*leavec;
          x:= 3/4;
          coastm:= x*protmp*coast*income*sea/bases;
          x:= 3/4;
          dryld:= x*drymp*dryl*income*sea/bases/60;
          IF ds > 0 THEN BEGIN
             x:= 3/4;
             coastc:= (1-x)*0.5*protmp*coast*income*(ds/dbss + sqr(ds/dbss))/60
                      +rhocp*coastc;
             x:= 3/4;
             drylt:=  (1-x)*0.5*drymp*dryl*income*(ds/dbss + sqr(ds/dbss))/60
                      +rhodl*drylt;
             x:= 0;
             welti:=  (1-x)*0.5*wetmp*wetl*income*(ds/dbss + sqr(ds/dbss))/60*ic
                      + rhowi*welti;
             weltt:=  x*0.5*wetmp*wetl*income*(ds/dbss + sqr(ds/dbss))/60
                      + rhowt*weltt;
          END
          ELSE BEGIN
               x:= 3/4;
               coastc:= 0.2*(1-x)*0.5*protmp*coast*income*(-ds/dbss + sqr(ds/dbss))/60
                        + rhocp*coastc;
               x:= 3/4;
               drylt:=  0.2*(1-x)*0.5*drymp*dryl*income*(-ds/dbss + sqr(ds/dbss))/60
                        + rhodl*drylt;
               x:= 0;
               welti:=  (1-x)*0.2*0.5*wetmp*wetl*income*(-ds/dbss + sqr(ds/dbss))/60*ic
                        + rhowi*welti;
               weltt:=  x*0.2*0.5*wetmp*wetl*income*(-ds/dbss + sqr(ds/dbss))/60
                        + rhowt*weltt;
          END;
          eloss:= coastm + coastc + dryld + drylt + weltt + enterc
                  + agri1 + agri2 + hurr2 + hurr4 + flooddam + winddam;
          sloss:= spec + stress + malcost + welti + leavec + hurr5;
(*          IF (name <> 'OECD-A') AND (name <> 'OECD-E') AND (name <> 'OECD-P') THEN BEGIN
             excessloss:= eloss + sloss - compth*income;
             IF excessloss > 0 THEN BEGIN
                compens:= 0;
                IF economy[1].income > excessloss*0.52 THEN BEGIN
                   economy[1].compens:= economy[1].compens-excessloss*0.52;
                   compens:= compens + 0.52*excessloss
                END;
                IF economy[2].income > excessloss*0.41 THEN BEGIN
                   economy[2].compens:= economy[2].compens-excessloss*0.41;
                   compens:= compens + 0.41*excessloss
                END;
                IF economy[3].income > excessloss*0.07 THEN BEGIN
                   economy[3].compens:= economy[3].compens-excessloss*0.07;
                   compens:= compens + excessloss*0.07
                END
             END
          END; *)
          income:= income - formp*forcost*y;
          IF i >= 40 THEN
             population:= population - leav + enter - dead;
          pg:= (population-popold)/popold;
          welfare:= (1-consleak)*income - eloss (*- apemp*edamairpo*);
          IF i >= 40 THEN
             income:= income - consleak*(eloss (*+ apemp*edamairpo*));
          IF income < 0 THEN
             income:= 0;
          welfare:= log(1000*welfare/population);
          IF income < 0.00000001 THEN
             welfare:= -100000000
          ELSE
              welfare:= welfare - sloss/income - 0.5*quad(sloss/income)
                        (*- apsmp*sdamairpo/income
                        - 0.5*quad(apsmp*sdamairpo/income)*);
          IF t = 40+10*optper THEN BEGIN
             cdr:= 1;
             hNPVW:= welfare
          END;
          IF (t > 40+10*optper) (*AND (t <= 50 + 10*optper)*) THEN BEGIN
             cdr:= (1+discrate)*cdr (*pow(1+discrate,(t-40-10*optper))*);
             IF hNPVW >= 1.0E-309 THEN
                hNPVW:= hNPVW + welfare/cdr
          END;
(*          IF (t > 50+10*optper) THEN BEGIN
             cdr:= pow(1+discrate,11);
             IF hNPVW >= 1.0E-309 THEN
                hNPVW:= hNPVW + welfare/cdr
          END;   *)
          totalpop:= totalpop + population
     END
END;

PROCEDURE Write1(VAR country: regrec; t: word; VAR els,sls,
                     (*eap,sap,*)ent,die,lea,inc,wel,pop,marg: text);

BEGIN
     WITH country, climate DO BEGIN
          write(ent,enter/population*1000000:10:6,' ');
          write(lea,leav/population*1000000:10:6,' ');
          write(die,dead/population*1000000:10:6,' ');
          IF income > 0.000001 THEN BEGIN
(*             write(eap,edamairpo/income*100:10:6,' ');
             write(sap,sdamairpo/income*100:10:6,' '); *)
             write(els,eloss/income*100:10:6,' ');
             write(sls,sloss/income*100:10:6,' ')
          END
          ELSE BEGIN
               (*write(eap,0,' ');
               write(sap,0,' ');*)
               write(els,0,' ');
               write(sls,0,' ')
          END;
          write(inc,income:12:6,' ');
          write(wel,emission:12:6,' ');
          IF t >+ 40 THEN
             write(marg,2*tax[trunc((t-40)/10)]*emission*income/1000:12:6,' ')
          ELSE
              write(marg,'    0.000000 ');
          write(pop,population:12:6,' ');
     END
END;

PROCEDURE CalcClimate(j: word);
(* calculates the climatic variables at time J *)

VAR mco2: double;
    i: word;
    ht, hdt: double;

BEGIN
     WITH climate DO BEGIN
          mco2:= 0;
          FOR i:= 1 TO NoReg DO
              WITH economy[i] DO BEGIN
                   mco2:= mco2 + emission + foremp*forestemm^[j] - forestry[trunc(j/10)]
              END;
          FOR i:= 1 TO 5 DO
              cbox[i]:= cbox[i]*decay[i] + 0.000471*frac[i]*(mco2 + addco2^[j]);
          acco2^[j]:= 0;
          FOR i:= 1 TO 5 DO
              acco2^[j]:= acco2^[j] + cbox[i];
          acch4^[j]:= acch4^[j-1] + 0.3597*(ch4em^[j] + addch4^[j])
                         - ch4decay*(acch4^[j-1]-ch4pre);
          acn2o^[j]:= acn2o^[j-1] + 0.2079*(n2oem^[j] + addn2o^[j])
                         - n2odecay*(acn2o^[j-1]-n2opre);

          rfCO2:= 6.3*log(acco2^[j]/co2pre);
          rfCH4:= 0.036*(sqrt(acCH4^[j])-sqrt(ch4pre))
                       - interact(acch4^[j],n2opre) + CH4N2O;
          rfN2O:= 0.14*(sqrt(acN2O^[j])-sqrt(N2Opre))
                       - interact(ch4pre,acn2o^[j]) + CH4N2O;

          radforc:= rfCO2 + rfCH4 + rfN2O;
          dt:= delaytemp*temps*radforc - delaytemp*temp;
          temp:= dt + temp;
          IF scengmt THEN BEGIN
             temp:= scentemp^[j];
             dt:= scentemp^[j] - scentemp^[j-1]
          END;
          flood:= floods*temp;
          wind:= winds*temp;
          tsq:= quad(temp);
          dtsq:= quad(dt);
          ds:= delaysea*seas*temp - delaysea*sea;
          sea:= ds + sea;
          ssq:= quad(sea);
          dssq:= quad(ds);
          hurr:= hurrs*temp
     END
END;

PROCEDURE DrawRandom;

VAR i, j: word;
    hulp, fake: double;

BEGIN
    WITH climate DO BEGIN
         GenGamma(8.1270,0.3508,temps); (*mode 2.5; var 1*)
(*         GenGamma(4.9218,0.6375,temps); (*mode 2.5; var 2*)
(*         GenGamma(4.27,1.16,temps);       (*mode 3.8; sd 2.4*)
         temps:= temps/6.3/ln(2);
(*         GenGamma(6.3584,0.0317,seas); (* mode 0.17; var 0.08^2 *)
(*         GenGamma(4.0083,0.0565,seas); (* mode 0.17; var 2*0.08^2 *)
(*         GenGamma(7.4729,0.0263,seas);      (* mode 0.17; sd 0.07 *)
         GenGamma(5.9959077,0.0612581,seas); (* mode 0.30604; var 0.15^2 *)
         GenNormal(modeh,0.10,hurrs,fake);
         GenNormal(0,1,floods,winds);
         floods:= modef+modef*floods;
         winds:= modew+modew*winds;
         GenTriang(8,16,1/ch4dec,hulp);
         ch4decay:= 1/hulp;
         GenTriang(100,170,1/n2odec,hulp);
         n2odecay:= 1/hulp;
         GenTriang(0.5*lifetemp,2*lifetemp,lifetemp,hulp);
         delaytemp:= 1/hulp;
         GenTriang(0.5*lifesea,2*lifesea,lifesea,hulp);
         delaysea:= 1/hulp;
         GenNormGTZ(lifeco2,0.25*lifeco2,hulp,fake);
         decay[2]:= exp(-1/hulp);
         GenNormGTZ(lifeco3,0.25*lifeco3,hulp,fake);
         decay[3]:= exp(-1/hulp);
         GenNormGTZ(lifeco4,0.25*lifeco4,hulp,fake);
         decay[4]:= exp(-1/hulp);
         GenNormGTZ(lifeco5,0.25*lifeco5,hulp,fake);
         decay[5]:= exp(-1/hulp);
    END;
    FOR i:= 1 TO NoReg DO
        WITH economy[i] DO BEGIN
             gennormal(1,0.2,ecgrmp,popgrmp);
             gennormal(1,0.2,aeeimp,aceimp);
             gennormal(1,0.2,foremp,fake);

             gennormal(1,0.5,specmp,heatmp);
             gennormal(1,0.5,coldmp,malarmp);
             gennormal(1,0.5,agrimp,hurrlmp);
             gennormal(1,0.5,hurrpmp,floodmp);
             gennormal(1,0.5,windmp,migmp);
             gennormal(1,0.5,protmp,drymp);
             gennormal(1,0.5,wetmp,fake);

             gengamma(2,1,voslmp);

(*             gengamma(2.6180,0.6180,specmp);
             gengamma(2.6180,0.6180,heatmp);
             gengamma(2.6180,0.6180,coldmp);
             gengamma(2.6180,0.6180,malarmp);
             gengamma(2.6180,0.6180,hurrlmp);
             gengamma(2.6180,0.6180,hurrpmp);
             gengamma(2.6180,0.6180,floodmp);
             gengamma(2.6180,0.6180,windmp);
             gengamma(2.6180,0.6180,migmp);
             gengamma(2.6180,0.6180,protmp);
             gengamma(2.6180,0.6180,drymp);
             gengamma(2.6180,0.6180,wetmp);
             gennormal(1,1,agrimp,fake); *)

             gennormGTZ(100,25,rhos,fake);
             rhos:= 1 - 1/rhos;
             gennormGTZ(10,2.5,rhoa,rhowt);
             rhoa:= 1 - 1/rhoa;
             rhowt:= 1 - 1/rhowt;
             gennormGTZ(15,3.75,rhohc,fake);
             rhohc:= 1 - 1/rhohc;
             gennormGTZ(5,1.25,rhoh,rhoi);
             rhoh:= 1 - 1/rhoh;
             rhoi:= 1 - 1/rhoi;
             gennormGTZ(5,1.25,rhoe,fake);
             rhoe:= 1 - 1/rhoe;
             gennormGTZ(50,12.5,rhowi,rhodl);
             rhowi:= 1 - 1/rhowi;
             rhodl:= 1 - 1/rhodl;
             gennormGTZ(50,12.5,rhocp,fake);
             rhocp:= 1 - 1/rhocp;

             gengamma(5.8284,0.2071,taxmp);
             gengamma(5.8284,0.2071,formp);
(*             gengamma(5.8284,0.2071,cregmp);  (* var = 0.25, mode = 1 *)
(*             gengamma(9.8990,0.1124,taxmp);
             gengamma(9.8990,0.1124,formp);
             gengamma(9.8990,0.1124,cregmp);   (* var = 0.125, mode = 1 *)
        END
END;

PROCEDURE RunEconomy;
(* calculates the economic and climatic variables for the entire period *)

VAR i, j, run : word;
    worldemm: double;

BEGIN
(*write('.');*)
     IF uncertain THEN BEGIN
        randseed:= 34514325;
        FOR run:= 1 TO runs DO BEGIN
(*gotoxy(3,23);write(run:5);*)
            DrawRandom;
            ReInit;
            FOR i:= 1 TO NYear DO BEGIN
                totalpop:= 0;
                FOR j:= 1 TO NoReg DO
                    Calculate(economy[j],i);
                IF i = 49+10*optper THEN
                   FOR j:= 1 TO NoReg DO BEGIN
                       emmrec[j]:= emmrec[j]*(run-1)/run + economy[j].emission/run;
                       IF economy[j].emission > ncemmres[j,optper] THEN
                          count[j]:= count[j]+1
                   END;
                FOR j:= 1 TO NoReg DO
                    WITH economy[j] DO BEGIN
                         income:= income + compens;
                         compens:= 0
                    END;
                CalcClimate(i)
            END;
            FOR i:= 1 TO NoReg DO
                economy[i].NPVW:= economy[i].hNPVW/run +
                                     (run-1)*economy[i].NPVW/run
        END
     END
     ELSE BEGIN
          ReInit;
          FOR i:= 1 TO NYear DO BEGIN
              totalpop:= 0;
              FOR j:= 1 TO NoReg DO
                  Calculate(economy[j],i);
              IF i = 49+10*optper THEN BEGIN
                 worldemm:= 0;
                 FOR j:= 1 TO NoReg DO BEGIN
                     worldemm:= worldemm + economy[j].emission (*+
                                economy[j].foremp*economy[j].forestemm^[i] -
                                economy[j].forestry[trunc(i/10)]*);
                     emmrec[j]:= economy[j].emission
                 END;
                 globviol:= (worldemm > cemmres[optper])
              END;
              FOR j:= 1 TO NoReg DO
                  WITH economy[j] DO BEGIN
                       income:= income + compens;
                       compens:= 0
                  END;
              CalcClimate(i)
          END;
          C2200:= climate.acco2^[NYear];
          FOR i:= 1 TO NoReg DO
(*              IF C2200 > CMax THEN
                 economy[i].NPVW:= economy[i].hNPVW*
                                   (1-0.5*(noper-optper)/noper*(C2200-CMax)*(C2200-CMax)/(CBaU-CMax)/(CBaU-CMax))
              ELSE *)
                  economy[i].NPVW:= economy[i].hNPVW
     END
END;

PROCEDURE WriteEMF;

VAR i, j, k, run : word;
    oecd,noecd,clim:text;
    TCEm, TCFE, TInc, TCCC, TERC: double;

BEGIN
     ReInit;
     assign(oecd,sessionpathname+'oecd.prn');
     rewrite(oecd);
     assign(noecd,sessionpathname+'noecd.prn');
     rewrite(noecd);
     assign(clim,sessionpathname+'climate.prn');
     rewrite(clim);
     FOR i:= 1 TO NYear DO BEGIN
         totalpop:= 0;
         FOR j:= 1 TO NoReg DO
             Calculate(economy[j],i);
             FOR k:= 1 TO NoReg DO
                 WITH economy[k] DO BEGIN
                      income:= income + compens;
                      compens:= 0
                 END;
         CalcClimate(i);
         IF (*(i mod 10 = 0)*) 0=0 THEN BEGIN
            TCEm:= 0;
            TCFE:= 0;
            TInc:= 0;
            TCCC:= 0;
            TERC:= 0;
            FOR j:= 1 TO 3 DO
                WITH economy[j] DO BEGIN
                     TCEm:= TCEm + emission;
                     TCFE:= TCFE + foremp*forestemm^[i] - forestry[trunc(i/10)];
                     TInc:= TInc + income;
                     TCCC:= TCCC + eloss + sloss;
                     TERC:= TERC + 0.01*ryg*income + formp*forcost*y
                END;
            writeln(oecd,TCEm:12:6,' ',TCFE:12:6,' ',TInc:12:6,' ',TCCC:12:6,' ',TERC:12:6);
            TCEm:= 0;
            TCFE:= 0;
            TInc:= 0;
            TCCC:= 0;
            TERC:= 0;
            FOR j:= 4 TO NoReg DO
                WITH economy[j] DO BEGIN
                     TCEm:= TCEm + emission;
                     TCFE:= TCFE + foremp*forestemm^[i] - forestry[trunc(i/10)];
                     TInc:= TInc + income;
                     TCCC:= TCCC + eloss + sloss;
                     TERC:= TERC + 0.01*ryg*income + formp*forcost*y
                END;
            writeln(noecd,TCEm:12:6,' ',TCFE:12:6,' ',TInc:12:6,' ',TCCC:12:6,' ',TERC:12:6);
            WITH Climate DO BEGIN
                 write(clim,i+1950:4,' ',acco2^[i]:10:6,' ',acch4^[i]:10:6,' ',acn2o^[i]:10:6,' ',temp:10:6,' ');
                 writeln(clim,sea:12:6,' ',hurr:12:6,' ',wind:12:6,' ',flood:12:6)
            END
         END;
     END;
     close(oecd);
     close(noecd);
     close(clim)
END;

PROCEDURE WriteTol;

VAR i, j, k, run : word;
    econ,clim:text;
    TCEm, TCFE, TInc, TCCC, TERC: double;

BEGIN
     ReInit;
     assign(econ,sessionpathname+'tolecon.prn');
     rewrite(econ);
     assign(clim,sessionpathname+'tolclim.prn');
     rewrite(clim);
     FOR j:= 1 TO NoReg DO
         WITH economy[j] DO BEGIN
              TCFE:= foremp*forestemm^[0];
              TERC:= 0.01*ryg*income + forcost;
              IF income > 0.0000001 THEN BEGIN
              write(econ,1950:4,' ',emission:12:6,' ',TCFE:12:6,' ',income:12:6,' ',population:12:6,' ',population/pop0:10:6);
              write(econ,income/population*1000:14:6,' ',eloss:12:6,' ',eloss/income*100:12:6,' ',sloss:12:6,' ');
              writeln(econ,sloss/income*100:12:6,' ',TERC:12:6,' ',terc/income*100:12:6,' ',name)
              END
              ELSE BEGIN
              write(econ,1950:4,' ',emission:12:6,' ',TCFE:12:6,' ',income:12:6,' ',population:12:6,' ',population/pop0:10:6);
              write(econ,income/population*1000:14:6,' ',eloss:12:6,' ',0,' ',sloss:12:6,' ');
              writeln(econ,0,' ',TERC:12:6,' ',0,' ',name)
              END
         END;
     WITH Climate DO BEGIN
          write(clim,1950:4,' ',acco2^[20]:12:6,' ',temp:12:6,' ',sea:8:6,' ',hurr:8:6,' ');
          writeln(clim,wind:8:6,' ',flood:8:6)
     END;
     FOR i:= 1 TO NYear DO BEGIN
         totalpop:= 0;
         FOR j:= 1 TO NoReg DO
             Calculate(economy[j],i);
         FOR k:= 1 TO NoReg DO
             WITH economy[k] DO BEGIN
                  income:= income + compens;
                  compens:= 0
             END;
         CalcClimate(i);
         IF (i=40) OR (i=50) OR (i=75) OR (i=100) OR (i=150) OR (i=250) THEN BEGIN
            FOR j:= 1 TO NoReg DO
                WITH economy[j] DO BEGIN
                     TCFE:= foremp*forestemm^[i] - forestry[trunc(i/10)];
                     TInc:= TInc + income;
                     TERC:= 0.01*ryg*income + formp*forcost*y;
IF income > 0.00000001 THEN BEGIN
write(econ,i+1950:4,' ',emission:12:6,' ',TCFE:12:6,' ',income:12:6,' ',population:12:6,' ',population/pop0:10:6);
write(econ,income/population*1000:14:6,' ',eloss:12:6,' ',eloss/income*100:12:6,' ',sloss:12:6,' ');
writeln(econ,sloss/income*100:12:6,' ',TERC:12:6,' ',terc/income*100:12:6,' ',name)
END
ELSE BEGIN
write(econ,i+1950:4,' ',emission:12:6,' ',TCFE:12:6,' ',income:12:6,' ',population:12:6,' ',population/pop0:10:6);
write(econ,income/population*1000:14:6,' ',eloss:12:6,' ',0,' ',sloss:12:6,' ');
writeln(econ,0,' ',TERC:12:6,' ',0,' ',name)
END
                END;
            WITH Climate DO BEGIN
                 write(clim,i+1950:4,' ',acco2^[i]:12:6,' ',temp:12:6,' ',sea:12:6,' ',hurr:12:6,' ');
                 writeln(clim,wind:12:6,' ',flood:12:6)
            END
         END;
     END;
     close(econ);
     close(clim)
END;

PROCEDURE RunEconomy1;
(* calculates the economic and climatic variables for the entire period *)

VAR i, j, k, run : word;
    els,sls,eap,sap,ent,die,lea,inc,wel,pop,marg,clim:text;
    wunc,cunc,tunc,punc,gunc,iunc,eunc,runc: text;
    totpop,totinc,totimp,totemm,totred: double;
    hp,hy,hw: regarr;

BEGIN
     IF uncertain THEN BEGIN
        randseed:= 34514325;
        assign(wunc,sessionpathname+'welfunc.prn');
        rewrite(wunc);
        assign(cunc,sessionpathname+'concunc.prn');
        rewrite(cunc);
        assign(tunc,sessionpathname+'tempunc.prn');
        rewrite(tunc);
        assign(punc,sessionpathname+'popunc.prn');
        rewrite(punc);
        assign(gunc,sessionpathname+'incunc.prn');
        rewrite(gunc);
        assign(iunc,sessionpathname+'impunc.prn');
        rewrite(iunc);
        assign(eunc,sessionpathname+'emmunc.prn');
        rewrite(eunc);
        assign(runc,sessionpathname+'redunc.prn');
        rewrite(runc);
        FOR run:= 1 TO runs DO BEGIN
gotoxy(3,23);write(run:5);
            DrawRandom;
            ReInit;
            FOR i:= 1 TO NYear DO BEGIN
                totalpop:= 0;
                FOR j:= 1 TO NoReg DO
                    Calculate(economy[j],i);
                FOR j:= 1 TO NoReg DO
                    WITH economy[j] DO BEGIN
                         income:= income + compens;
                         compens:= 0
                    END;
                CalcClimate(i);
                IF (i=50) OR (i=75) OR (i=100) OR (i=125) OR (i=150) OR (i=NYear) THEN BEGIN
                   write(tunc,climate.temp:8:4,' ');
                   totpop:= 0;
                   totinc:= 0;
                   totimp:= 0;
                   totemm:= 0;
                   totred:= 0;
                   FOR k:= 1 TO NoReg DO BEGIN
                       totpop:= totpop + economy[k].population;
                       totinc:= totinc + economy[k].income;
                       totemm:= totemm + economy[k].emission;
                       totimp:= totimp + economy[k].eloss + economy[k].sloss;
                       totred:= totred + economy[k].taxmp*economy[k].ryg*economy[k].income;
                   END;
                   write(punc,totpop:10:4,' ');
                   write(gunc,totinc:10:4,' ');
                   write(eunc,totemm:10:4,' ');
                   write(iunc,totimp/totinc*100:8:4,' ');
                   write(runc,totred/totinc*100:10:8,' ')
                END
            END;
            writeln(tunc);
            writeln(punc);
            writeln(gunc);
            writeln(eunc);
            writeln(iunc);
            writeln(runc);
            FOR i:= 1 TO NoReg DO
                WITH Economy[i] DO BEGIN
                     NPVW:= hNPVW/run + (run-1)*NPVW/run;
                     write(wunc,hNPVW:20:4,' ')
                END;
            writeln(wunc);
            WITH climate DO BEGIN
                 c2200:= acco2^[NYear]/run + (run-1)*c2200/run;
                 c2100:= acco2^[NYear-100]/run + (run-1)*c2100/run;
                 c2075:= acco2^[NYear-125]/run + (run-1)*c2075/run;
                 c2050:= acco2^[NYear-150]/run + (run-1)*c2050/run;
                 c2025:= acco2^[NYear-175]/run + (run-1)*c2025/run;
                 c2000:= acco2^[NYear-200]/run + (run-1)*c2000/run;
                 write(cunc,acco2^[NYear-200]:10:4,' ',acco2^[NYear-175]:10:4,' ');
                 write(cunc,acco2^[NYear-150]:10:4,' ',acco2^[NYear-125]:10:4,' ');
                 writeln(cunc,acco2^[NYear-100]:10:4,' ',acco2^[NYear]:10:4)
            END
        END;
        close(wunc);
        close(cunc);
        close(tunc);
        close(punc);
        close(eunc);
        close(gunc);
        close(iunc);
        close(runc);
        FOR i:= 1 TO NoReg DO
            WITH Economy[i] DO BEGIN
                 p:= hp[i];
                 w1950:= hw[i];
                 y:= hy[i]
            END
     END
     ELSE BEGIN
          ReInit;
          assign(els,sessionpathname+'eloss.prn');
          rewrite(els);
          assign(sls,sessionpathname+'sloss.prn');
          rewrite(sls);
          assign(ent,sessionpathname+'enter.prn');
          rewrite(ent);
          assign(die,sessionpathname+'dead.prn');
          rewrite(die);
          assign(lea,sessionpathname+'leave.prn');
          rewrite(lea);
          assign(inc,sessionpathname+'income.prn');
          rewrite(inc);
          assign(wel,sessionpathname+'emissions.prn');
          rewrite(wel);
          assign(pop,sessionpathname+'population.prn');
          rewrite(pop);
          assign(marg,sessionpathname+'margcost.prn');
          rewrite(marg);
          assign(clim,sessionpathname+'climate.prn');
          rewrite(clim);
          FOR i:= 1 TO NYear DO BEGIN
              totalpop:= 0;
              write(els,i+1950,' ');
              write(sls,i+1950,' ');
              write(ent,i+1950,' ');
              write(die,i+1950,' ');
              write(lea,i+1950,' ');
              write(inc,i+1950,' ');
              write(wel,i+1950,' ');
              write(marg,i+1950,' ');
              write(pop,i+1950,' ');
              FOR j:= 1 TO NoReg DO BEGIN
                  Calculate(economy[j],i);
                  FOR k:= 1 TO NoReg DO
                      WITH economy[k] DO BEGIN
                           income:= income + compens;
                           compens:= 0
                      END;
                  Write1(economy[j],i,els,sls,ent,die,lea,inc,wel,pop,marg)
              END;
              CalcClimate(i);
              WITH climate DO BEGIN
                   write(clim,i+1950:4,' ',acco2^[i]:10:6,' ',acch4^[i]:10:6,' ',acn2o^[i]:10:6,' ',temp:10:6,' ');
                   writeln(clim,sea:8:6,' ',flood:8:6,' ',wind:8:6,' ',hurr:8:6)
              END;
              writeln(els);
              writeln(sls);
              writeln(ent);
              writeln(die);
              writeln(lea);
              writeln(inc);
              writeln(wel);
              writeln(marg);
              writeln(pop)
          END;
          close(els);
          close(sls);
          close(ent);
          close(die);
          close(lea);
          close(inc);
          close(wel);
          close(marg);
          close(pop);
          close(clim);
          WriteTol;
          FOR i:= 1 TO NoReg DO
              economy[i].NPVW:= economy[i].hNPVW
     END
END;

PROCEDURE ShortRep(name: string);

VAR i: word;
    outf: text;

BEGIN
     assign(outf,sessionpathname+name);
     rewrite(outf);
     WriteLn(outf,'     tax   reg    forestry  region');
     FOR i:= 1 TO NoReg DO
         WITH economy[i] DO
              WriteLn(outf,tax[optper]:8:4,' ',forestry[optper]:8:2,'  ',name);
     WriteLn(outf);
     WriteLn(outf,'            NPVW        p        y        w       v    region');
     FOR i:= 1 TO NoReg DO
         WITH economy[i] DO
              WriteLn(outf,NPVW:16:4,' ',p:8:4,' ',y:8:4,' ',w1950:8:4,' ',v:8:4,' ',name);
     WriteLn(outf);
     WriteLn(outf,'  income 90   captrans  region');
     FOR i:= 1 TO NoReg DO
         WITH economy[i] DO
              WriteLn(outf,inc0:11:6,' ',captrans:10:6,'  ',name);
     WriteLn(outf);
     FOR i:= 28 TO NYear DO
         IF ((i+1970) MOD 25) = 0 THEN
            Writeln(outf,climate.acco2^[i]:12:6);
     close(outf)
END;

FUNCTION Objective: double;
(* calculates the sum total of the regional NPWs *)

VAR j: word;
    objw: double;

BEGIN
     objw:= 0;
     FOR j:= 1 TO NoReg DO
         objw:= objw + economy[j].npvw;
     objective:= objw
END;

(* OPTIMISING PROCEDURES *)

PROCEDURE CoopCBAction(VAR T,F: measure);
(* finds the optimal emission restriction by taxing (OECD) or energy and
   carbon efficiency (non-OECD) and forestry measures under cooperation *)

VAR i, rc, mc: word;
    obj0, hobj, mobj: double;

BEGIN
     RunEconomy;
     obj0:= objective;
(*write(obj0:8:4,' ');*)
     mobj:= -10000000;
     FOR i:= 1 TO NoReg DO BEGIN
         objtax(i);
         RunEconomy;
         hobj:= objective;
         IF (hobj > mobj) THEN BEGIN
            mobj:= hobj;
            rc:= i;
(*write('tax ',economy[i].name,' ');*)
            mc:= 1;
         END;
         taxobj(i);

         objfor(i);
         RunEconomy;
         hobj:= objective;
         IF (hobj > mobj) THEN BEGIN
             mobj:= hobj;
(*write('for ',economy[i].name,' ');*)
             rc:= i;
             mc:= 3;
         END;
         forobj(i)
     END;
     WHILE mobj > obj0 DO BEGIN
           IF mc=1 THEN
              objtax(rc);
           IF mc=3 THEN
              objfor(rc);
           RunEconomy;
           obj0:= objective;
(*write(obj0:8:4,' ');*)
           mobj:= -10000000;
           FOR i:= 1 TO NoReg DO BEGIN
               objtax(i);
               RunEconomy;
               hobj:= objective;
               IF (hobj > mobj) THEN BEGIN
                  mobj:= hobj;
                  rc:= i;
(*write('tax ',economy[i].name,' ');*)
                  mc:= 1;
               END;
               taxobj(i);

               objfor(i);
               RunEconomy;
               hobj:= objective;
               IF (hobj > mobj) THEN BEGIN
                  mobj:= hobj;
(*write('for ',economy[i].name,' ');*)
                  rc:= i;
                  mc:= 3;
               END;
               forobj(i)
           END
     END;
     FOR i:= 1 TO NoReg DO BEGIN
         T[i]:= economy[i].tax[optper];
         F[i]:= economy[i].forestry[optper]
     END
END;

PROCEDURE CoopCEAction(psnr: word; VAR T,F: measure);

VAR i, j, rc, mc: word;
    obj0, hobj, mobj: double;

BEGIN
     RunEconomy;
     obj0:= objective;
(*write(obj0:8:4,' ');*)
     mobj:= -10000000;
     FOR i:= 1 TO psnr DO BEGIN
         objtax(i);
         RunEconomy;
         hobj:= objective;
         IF (hobj > mobj) THEN BEGIN
            mobj:= hobj;
            rc:= i;
(*write('tax ',economy[i].name,' ');*)
            mc:= 1;
         END;
         taxobj(i);

         objfor(i);
         RunEconomy;
         hobj:= objective;
         IF (hobj > mobj) THEN BEGIN
             mobj:= hobj;
(*write('for ',economy[i].name,' ');*)
             rc:= i;
             mc:= 3;
         END;
         forobj(i)
     END;
     WHILE (*(mobj > obj0) OR*) globviol DO BEGIN
           IF mc=1 THEN
              objtax(rc);
           IF mc=3 THEN
              objfor(rc);
           RunEconomy;
           obj0:= objective;
(*write(obj0:8:4,' ');*)
           mobj:= -10000000;
           FOR i:= 1 TO psnr DO BEGIN
               objtax(i);
               RunEconomy;
               hobj:= objective;
               IF (hobj > mobj) THEN BEGIN
                  mobj:= hobj;
                  rc:= i;
(*write('tax ',economy[i].name,' ');*)
                  mc:= 1;
               END;
               taxobj(i);

               objfor(i);
               RunEconomy;
               hobj:= objective;
               IF (hobj > mobj) THEN BEGIN
                  mobj:= hobj;
(*write('for ',economy[i].name,' ');*)
                  rc:= i;
                  mc:= 3;
               END;
               forobj(i)
           END
     END;
     FOR i:= 1 TO NoReg DO BEGIN
         T[i]:= economy[i].tax[optper];
         F[i]:= economy[i].forestry[optper]
     END
END;

PROCEDURE NonCoopCEAction(VAR T,F: measure);

VAR i, j: word;
    emm0, tobj, fobj, obj0: double;
    cont: boolean;

BEGIN
(*write(economy[i].name,' ');*)
(*     FOR i:= 1 TO NoReg DO
         count[i]:= 0;*)
     RunEconomy;
(*writeln(ncemmres[i,optper]:8:4,' ',emmrec[i]:8:4,' ',economy[i].tax[optper]:6:4);*)
         (*obj0:= economy[i].npvw;
         emm0:= emmrec[i];
         objtax(i);
         count[i]:= 0;
         RunEconomy;
         tobj:= economy[i].npvw;
         taxobj(i);
(*         objfor(i);
         RunEconomy;
         fobj:= economy[i].npvw;
         forobj(i);   *)
     cont:= false;
     FOR i:= 1 TO NoReg DO
         IF emmrec[i] > ncemmres[i,optper] (*count[i] > 0.5*runs*) THEN BEGIN
            objtax(i);
            cont:= true
         END;
     WHILE cont DO BEGIN
(*           FOR i:= 1 TO NoReg DO
               count[i]:= 0;*)
           RunEconomy;
           cont:= false;
           FOR i:= 1 TO NoReg DO
               IF emmrec[i] > ncemmres[i,optper] (*count[i] > 0.5*runs*) THEN BEGIN
                  objtax(i);
                  cont:= true
               END
     END;


         (*) (count[i] > 0.5*runs) (*OR (tobj > obj0) OR (fobj > obj0)*)(* DO BEGIN
              (* IF tobj > fobj THEN*)
              (*    objtax(i)
              (* ELSE
                   objfor(i);*)
(*writeln(ncemmres[i,optper]:8:4,' ',emmrec[i]:8:4,' ',economy[i].tax[optper]:6:4);*)
              (* RunEconomy;
               (*obj0:= economy[i].npvw;
               emm0:= emmrec[i];
               objtax(i);
               count[i]:= 0;
               RunEconomy;
               tobj:= economy[i].npvw;
               taxobj(i);
            (*   objfor(i);
               RunEconomy;
               fobj:= economy[i].npvw;
               forobj(i)  *)
        (* END; *)
     FOR i:= 1 TO NoReg DO BEGIN
         T[i]:= economy[i].tax[optper];
         F[i]:= 0 (*economy[i].forestry[optper]*)
     END
END;

PROCEDURE CoopCBA(VAR T,F: measure);

VAR i,j,k: word;

BEGIN
     taxstep:= 16*btaxst;
     forstep:= 16*bforst;
     CoopCBAction(T,F);
     FOR i:= 1 TO NoReg DO BEGIN
         IF economy[i].tax[optper] > 0 THEN
            taxobj(i);
         IF economy[i].forestry[optper] > 0 THEN
            forobj(i)
     END;
     taxstep:= 4*btaxst;
     forstep:= 4*bforst;
     CoopCBAction(T,F);
     FOR i:= 1 TO NoReg DO BEGIN
         IF economy[i].tax[optper] > 0 THEN
            taxobj(i);
         IF economy[i].forestry[optper] > 0 THEN
            forobj(i)
     END;
     taxstep:= btaxst;
     forstep:= bforst;
     CoopCBAction(T,F)
END;

PROCEDURE CoopCEA(psnr: word; VAR T,F: measure);

VAR i,j,k: word;

BEGIN
     taxstep:= 16*btaxst;
     forstep:= 16*bforst;
     CoopCEAction(psnr,T,F);
     FOR i:= 1 TO psnr DO BEGIN
         IF economy[i].tax[optper] > 0 THEN
            taxobj(i);
         IF economy[i].forestry[optper] > 0 THEN
            forobj(i)
     END;
     taxstep:= 4*btaxst;
     forstep:= 4*bforst;
     CoopCEAction(psnr,T,F);
     FOR i:= 1 TO psnr DO BEGIN
         IF economy[i].tax[optper] > 0 THEN
            taxobj(i);
         IF economy[i].forestry[optper] > 0 THEN
            forobj(i)
     END;
     taxstep:= btaxst;
     forstep:= bforst;
     CoopCEAction(psnr,T,F)
END;

PROCEDURE NonCoopCEA(VAR T,F: measure);

VAR i,j,k: word;

BEGIN
     taxstep:= 16*btaxst;
     forstep:= 16*bforst;
     NonCoopCEAction(T,F);
     FOR i:= 1 TO NoReg DO BEGIN
         IF economy[i].tax[optper] > 0 THEN
            taxobj(i);
         IF economy[i].forestry[optper] > 0 THEN
            forobj(i)
     END;
     taxstep:= 4*btaxst;
     forstep:= 4*bforst;
     NonCoopCEAction(T,F);
     FOR i:= 1 TO NoReg DO BEGIN
         IF economy[i].tax[optper] > 0 THEN
            taxobj(i);
         IF economy[i].forestry[optper] > 0 THEN
            forobj(i)
     END;
     taxstep:= btaxst;
     forstep:= bforst;
     NonCoopCEAction(T,F)
END;

PROCEDURE NCOptimiseTaxFor(W: integer; VAR T,F,ObF: double);
(* finds the optimal emission restriction by taxing and forestry under
   non-cooperation *)

VAR objt, objf, obj0: double;

BEGIN
write(' ',C2200:8:2,' ',economy[w].name);

     taxstep:= 16*btaxst;
     forstep:= 16*bforst;
     RunEconomy;
     obj0:= (1-altruism)*economy[w].npvw + altruism*objective;
     objfor(W);
     RunEconomy;
     objf:= (1-altruism)*economy[w].npvw + altruism*objective;
     forobj(W);
     objtax(W);
     RunEconomy;
     objt:= (1-altruism)*economy[w].npvw + altruism*objective;
     taxobj(W);
     WHILE ((objt > obj0) OR (objf > obj0)) DO BEGIN
           IF objt > objf THEN BEGIN
              objtax(W);
           END
           ELSE BEGIN
                objfor(W);
                RunEconomy
           END;
           obj0:= (1-altruism)*economy[w].npvw + altruism*objective;
           objfor(W);
           RunEconomy;
           objf:= (1-altruism)*economy[w].npvw + altruism*objective;
           forobj(W);
           objtax(W);
           RunEconomy;
           objt:= (1-altruism)*economy[w].npvw + altruism*objective;
           taxobj(W)
     END;
     IF economy[w].tax[optper] > 0 THEN
        taxobj(w);
     IF economy[w].forestry[optper] > 0 THEN
        forobj(w);

     taxstep:= 4*btaxst;
     forstep:= 4*bforst;
     RunEconomy;
     obj0:= (1-altruism)*economy[w].npvw + altruism*objective;
     objfor(W);
     RunEconomy;
     objf:= (1-altruism)*economy[w].npvw + altruism*objective;
     forobj(W);
     objtax(W);
     RunEconomy;
     objt:= (1-altruism)*economy[w].npvw + altruism*objective;
     taxobj(W);
     WHILE ((objt > obj0) OR (objf > obj0)) DO BEGIN
           IF objt > objf THEN BEGIN
              objtax(W);
           END
           ELSE BEGIN
                objfor(W);
                RunEconomy
           END;
           obj0:= (1-altruism)*economy[w].npvw + altruism*objective;
           objfor(W);
           RunEconomy;
           objf:= (1-altruism)*economy[w].npvw + altruism*objective;
           forobj(W);
           objtax(W);
           RunEconomy;
           objt:= (1-altruism)*economy[w].npvw + altruism*objective;
           taxobj(W)
     END;
     IF economy[w].tax[optper] > 0 THEN
        taxobj(w);
     IF economy[w].forestry[optper] > 0 THEN
        forobj(w);

     taxstep:= btaxst;
     forstep:= bforst;
     RunEconomy;
     obj0:= (1-altruism)*economy[w].npvw + altruism*objective;
     objfor(W);
     RunEconomy;
     objf:= (1-altruism)*economy[w].npvw + altruism*objective;
     forobj(W);
     objtax(W);
     RunEconomy;
     objt:= (1-altruism)*economy[w].npvw + altruism*objective;
     taxobj(W);
     WHILE ((objt > obj0) OR (objf > obj0)) DO BEGIN
           IF objt > objf THEN BEGIN
              objtax(W);
           END
           ELSE BEGIN
                objfor(W);
                RunEconomy
           END;
           obj0:= (1-altruism)*economy[w].npvw + altruism*objective;
           objfor(W);
           RunEconomy;
           objf:= (1-altruism)*economy[w].npvw + altruism*objective;
           forobj(W);
           objtax(W);
           RunEconomy;
           objt:= (1-altruism)*economy[w].npvw + altruism*objective;
           taxobj(W)
     END;

     T:= economy[w].tax[optper];
     F:= economy[w].forestry[optper];
     ObF:= obj0
END;

PROCEDURE NCOptimisation(W: integer; VAR T, F: double);

VAR O1: double;

BEGIN
     NCOptimiseTaxFor(W,T,F,O1)
END;

PROCEDURE Optimal(T,F: measure);

var i: word;

BEGIN
     SetTax4(T);
     SetFor4(F);
     RunEconomy1;
     ShortRep('summary.prn')
END;

(* FUND PROCEDURES *)

PROCEDURE NCRevRec(VAR trans: matmeas);

VAR i: word;
    h, h0: double;

BEGIN
(*write('offer ');*)
     FOR i:= 1 TO NoReg DO BEGIN
         WITH economy[i], climate DO BEGIN
              h0:= npvw;
              objtax(i);
              RunEconomy;
              h:= h0 - npvw;
              WHILE h < 0 DO BEGIN
(*write(name,' add taxes ');*)
                    h0:= npvw;
                    objtax(i);
                    RunEconomy;
                    h:= h0 - npvw
              END;
              taxobj(i);
              h:= inc0*(expo(h/88.62596)-1);
              trans[i,1]:= h;
              trans[i,2]:= -1;
              trans[i,3]:= -1;

              objfor(i);
              RunEconomy;
              h:= h0 - npvw;
              WHILE h < 0 DO BEGIN
(*write(name,' add for ');*)
                    h0:= npvw;
                    objfor(i);
                    RunEconomy;
                    h:= h0 - npvw
              END;
              forobj(i);
              h:= inc0*(expo(h/88.62596)-1);
              trans[i,3]:= h
         END;
     RunEconomy
     END
END;

FUNCTION NCRevInv(kaptr: double; k,l: word; VAR outf: text): boolean;

VAR i,m: word;
    w: ARRAY [1..NoReg] OF double;
    max: double;

BEGIN
     FOR i:= 1 TO NoReg DO BEGIN
         WITH economy[i], climate DO BEGIN
              IF kaptr < 0.25*inc0 THEN BEGIN
                 captrans:= captrans - kaptr;
                 economy[k].captrans:= economy[k].captrans + kaptr;
                 CASE l OF
                      1: objtax(k);
                      3: objfor(k)
                 END;
                 w[i]:= npvw;
                 RunEconomy;
                 w[i]:= npvw - w[i];
                 captrans:= captrans + kaptr;
                 economy[k].captrans:= economy[k].captrans - kaptr;
                 CASE l OF
                      1: taxobj(k);
                      3: forobj(k)
                 END;
                 RunEconomy
              END
              ELSE
                  w[i]:= -10000
         END
     END;
     max:= w[1];
     m:= 1;
     FOR i:= 2 TO NoReg DO
         IF w[i] > max THEN BEGIN
            m:= i;
            max:= w[i]
         END;
     IF w[m] > 0 THEN BEGIN
(*write(' from ',economy[m].name,' to ',economy[k].name);
writeLn(outf,'from ',economy[m].name,' to ',economy[k].name,' for ',l:1,' a ',kaptr);*)
        economy[m].captrans:= economy[m].captrans - kaptr;
        economy[k].captrans:= economy[k].captrans + kaptr;
        CASE l OF
             1: objtax(k);
             3: objfor(k)
        END;
        RunEconomy;
        ncrevinv:= true
     END
     ELSE
         ncrevinv:= false
END;

PROCEDURE NCFund;

VAR trans: matmeas;
    l, k, i, j: word;
    cont: boolean;
    outf: text;

BEGIN
     cont:= true;
     assign(outf,sessionpathname+'captrans.prn');
     rewrite(outf);
     WHILE cont=true DO BEGIN
           NCRevRec(trans);
           cont:= false;
(*write('accept ');*)
           FOR i:= 1 TO NoReg DO
               FOR j:= 1 TO NOI DO
                   IF trans[i,j] >= 0 THEN BEGIN
                      cont:= NCRevInv(trans[i,j],i,j,outf) OR cont
                   END
     END;
     close(outf);
     RunEconomy1;
     shortrep('fundsumm.prn');
END;

FUNCTION IdentWinLoos(VAR h: measure): boolean;

VAR i: word;
    hinv, hrec: double;
    bh: ARRAY [1..NoReg] OF boolean;

BEGIN
     FOR i:= 1 TO NoReg DO
         WITH economy[i] DO BEGIN
              h[i]:= npvw - basewf; (*npvw0[i];*)
              bh[i]:= (h[i] > 0);
              h[i]:= inc0*(expo(h[i]/88.62596)-1)
         END;
     hinv:= 0;
     hrec:= 0;
     FOR i:= 1 TO NoReg DO
         IF bh[i] THEN
(*            IF h[i] < 0.005*economy[i].inc0 THEN *)
               hinv:= hinv + h[i]
(*            ELSE BEGIN
                h[i]:= 0.005*economy[i].inc0;
                hinv:= hinv + h[i]
            END*)
         ELSE
             hrec:= hrec + h[i];
     IF hinv >= abs(hrec) THEN BEGIN
        FOR i:= 1 TO NoReg DO
            IF bh[i] THEN
               h[i]:= h[i]*hrec/hinv
            ELSE
                h[i]:= - h[i];
        IdentWinLoos:= true
     END
     ELSE
         IdentWinLoos:= false;
END;

PROCEDURE COptimiseTaxFor(VAR T,F: measure);

VAR i, j, rc, mc: word;
    obj0, hobj, mobj: double;
    kt, hkt: measure;

BEGIN
     FOR i:= 1 TO NoReg DO BEGIN
         hkt[i]:= 0;
         kt[i]:= 0
     END;
     RunEconomy;
     obj0:= objective;
(*write(obj0:8:4,' ');*)
     mobj:= -10000000;
     FOR i:= 1 TO NoReg DO BEGIN
         objtax(i);
         RunEconomy;
         hobj:= objective;
         IF (hobj > mobj) AND IdentWinLoos(hkt) THEN BEGIN
            mobj:= hobj;
            rc:= i;
(*write('tax ',economy[i].name,' ',mobj:8:4);*)

            mc:= 1;
            FOR j:= 1 TO NoReg DO
                kt[j]:= hkt[j];
         END;
         taxobj(i);

         objfor(i);
         RunEconomy;
         hobj:= objective;
         IF (hobj > mobj) AND IdentWinLoos(hkt) THEN BEGIN
            mobj:= hobj;
            rc:= i;
(*write('for ',economy[i].name,' ');*)
            mc:= 3;
            FOR j:= 1 TO NoReg DO
                kt[j]:= hkt[j];
         END;
         forobj(i);
     END;
     WHILE mobj > obj0 DO BEGIN
           obj0:= mobj;
(*write(obj0:8:4);*)
           mobj:= -10000000;
           IF mc=1 THEN
              objtax(rc);
           IF mc=3 THEN
              objfor(rc);
           FOR j:= 1 TO NoReg DO BEGIN
               economy[j].captrans:= economy[j].captrans + kt[j];
               kt[j]:= 0;
               hkt[j]:= 0
           END;
           FOR i:= 1 TO NoReg DO BEGIN
               objtax(i);
               RunEconomy;
               hobj:= objective;
               IF (hobj > mobj) AND IdentWinLoos(hkt) THEN BEGIN
                  mobj:= hobj;
                  rc:= i;
(*write('tax ',economy[i].name,' ',mobj:8:4);*)
                  mc:= 1;
                  FOR j:= 1 TO NoReg DO
                      kt[j]:= hkt[j];
               END;
               taxobj(i);

               objfor(i);
               RunEconomy;
               hobj:= objective;
               IF (hobj > mobj) AND IdentWinLoos(hkt) THEN BEGIN
                  mobj:= hobj;
                  rc:= i;
(*write('for ',economy[i].name,' ');*)
                  mc:= 3;
                  FOR j:= 1 TO NoReg DO
                      kt[j]:= hkt[j];
               END;
               forobj(i);
           END;
     END;

     FOR i:= 1 TO NoReg DO BEGIN
         T[i]:= economy[i].tax[optper];
         F[i]:= economy[i].forestry[optper]
     END
END;

PROCEDURE COptTaxFor(VAR T,F: measure);

VAR i: word;

BEGIN
     RunEconomy;
     FOR i:= 1 TO NoReg DO
         economy[i].basewf:= economy[i].npvw;
     COptimiseTaxFor(T,F);
END;

PROCEDURE CFUND;

VAR T,F: measure;

BEGIN
     WriteLn;
     COptTaxFor(T,F);
     Optimal(T,F)
END;


PROCEDURE FUND;

BEGIN
     IF cooperate THEN
        CFund
     ELSE
         NCFund
END;

(* EXIT PROCEDURE *)

PROCEDURE Exit;

VAR i: word;

BEGIN
     FOR i:= 1 TO NoReg DO
         WITH Economy[i] DO BEGIN
              dispose(ypcgrowth);
              dispose(pgrowth);
              dispose(urbanp);
              dispose(aeei);
              dispose(acei);
              dispose(forestemm)
         END;
     WITH Climate DO BEGIN
          dispose(addco2);
          dispose(addch4);
          dispose(addn2o);
          dispose(ch4em);
          dispose(n2oem);
          dispose(acco2);
          dispose(acch4);
          dispose(acn2o)
     END
END;

(* MAIN PROCEDURES *)

PROCEDURE RunNonCoopCBA;

VAR T,F: measure;
    w, cnt: integer;
    outs: text;
    stop: boolean;
    s: string;
    seet, seef: double;

BEGIN
     SetMeasures1;
     IF sequent THEN BEGIN
(*        assign(outs,sessionpathname+'seq0.prn');
        rewrite(outs);
        FOR optper:= 0 TO NoPer DO BEGIN
            write(optper:2,' ');
            FOR w:= 1 TO NoReg DO BEGIN
                NCOptimisation(w,T[w],F[w]);
                economy[w].tax[optper]:= 0
            END;
            FOR w:= 1 TO NoReg DO
                write(outs,T[w]:6:4,' ');
            writeln(outs);
            FOR w:= 1 TO NoReg DO
                write(outs,F[w]:6:2,' ');
            writeln(outs);
        END;
        close(outs); *)
        cnt:= 7;
        REPEAT
              inc(cnt);
              str(cnt,s);
              assign(outs,sessionpathname+'seq'+s+'.prn');
              rewrite(outs);
              SetMeasures3(cnt-1);
              FOR optper:= 0 TO NoPer DO BEGIN
                  write(optper:2,' ');
                  FOR w:= 1 TO NoReg DO
                      WITH economy[w] DO BEGIN
                           seet:= tax[optper];
                           seef:= forestry[optper];
                           tax[optper]:= 0;
                           forestry[optper]:= 0;
                           NCOptimisation(w,T[w],F[w]);
                           tax[optper]:= seet;
                           forestry[optper]:= seef;
                      END;
                  FOR w:= 1 TO NoReg DO
                      write(outs,((*0.9*economy[w].tax[optper]+0.1*)T[w]):6:4,' ');
                  writeln(outs);
                  FOR w:= 1 TO NoReg DO
                      write(outs,((*0.9*economy[w].forestry[optper]+0.1*)F[w]):6:2,' ');
                  writeln(outs);
              END;
              close(outs);
              stop:= checkstop(cnt)
        UNTIL stop
     END
     ELSE BEGIN
          optper:= 0;
          FOR w:= 1 TO NoReg DO
              NCOptimisation(w,T[w],F[w]);
          Optimal(T,F);
          IF captrans THEN
             NCFund
     END
END;

PROCEDURE RunNonCoopCEA;

VAR T,F: measure;
    i,j,cnt: word;
    s: string;
    ceaout: text;

BEGIN
(*     SetMeasures1;
     assign(ceaout,sessionpathname+'seq0.prn');
     rewrite(ceaout);
     FOR optper:= 0 TO NoPer DO BEGIN
         write(optper:2,' ');
         RunEconomy;
         NonCoopCEA(T,F);
         FOR i:= 1 TO NoReg DO
             write(ceaout,T[i]:6:4,' ');
         writeln(ceaout);
         FOR i:= 1 TO NoReg DO
             write(ceaout,F[i]:6:2,' ');
         writeln(ceaout)
     END;
     close(ceaout); *)
     randseed:= 34514325;
     FOR j:= 1 TO Runs DO BEGIN
(*         write(j:5);*)
         str(j,s);
         assign(ceaout,sessionpathname+'per'+s+'.prn');
         rewrite(ceaout);
         DrawRandom;
         SetMeasures1;
     FOR optper:= 0 TO NoPer DO BEGIN
write(optper:2,' ');
         RunEconomy;
         NonCoopCEA(T,F);
         FOR i:= 1 TO NoReg DO
             write(ceaout,T[i]:6:4,' ');
         writeln(ceaout)
     END;
     close(ceaout);
     END;

(*     cnt:= 0;
     REPEAT
           inc(cnt);
           str(cnt,s);
           assign(ceaout,sessionpathname+'seq'+s+'.prn');
           rewrite(ceaout);
           SetMeasures3(cnt-1);
           FOR optper:= 0 TO NoPer DO BEGIN
               write(optper:2,' ');
               RunEconomy;
               NonCoopCEA(T,F);
               FOR i:= 1 TO NoReg DO
                   write(ceaout,T[i]:6:4,' ');
               writeln(ceaout);
               FOR i:= 1 TO NoReg DO
                   write(ceaout,F[i]:6:2,' ');
               writeln(ceaout)
           END;
           close(ceaout);
     UNTIL checkstop(cnt)*)
END;

PROCEDURE RunCoopCBA;

VAR T,F,seet,seef: measure;
    i, cnt: word;
    outs: text;
    s: string;
    stop: boolean;

BEGIN
     IF Sequent THEN BEGIN
        assign(outs,sessionpathname+'seq0.prn');
        rewrite(outs);
        FOR optper:= 0 TO NoPer DO BEGIN
            SetMeasures1;
            write(optper:2,' ');
            RunEconomy;
            FOR i:= 1 TO NoReg DO
                economy[i].basewf:= economy[i].npvw;
            CoopCBA(T,F);
            FOR i:= 1 TO NoReg DO
                write(outs,T[i]:6:4,' ');
            writeln(outs);
            FOR i:= 1 TO NoReg DO
                write(outs,F[i]:6:2,' ');
            writeln(outs);
        END;
        close(outs);
        cnt:= 0;
        REPEAT
              inc(cnt);
              str(cnt,s);
              assign(outs,sessionpathname+'seq'+s+'.prn');
              rewrite(outs);
              SetMeasures3(cnt-1);
              FOR optper:= 0 TO NoPer DO BEGIN
                  write(optper:2,' ');
                  FOR i:= 1 TO NoReg DO BEGIN
                      seet[i]:= economy[i].tax[optper];
                      seef[i]:= economy[i].forestry[optper];
                      economy[i].tax[optper]:= 0;
                      economy[i].forestry[optper]:= 0
                  END;
                  RunEconomy;
                  FOR i:= 1 TO NoReg DO
                      economy[i].basewf:= economy[i].npvw;
                  CoopCBA(T,F);
                  FOR i:= 1 TO NoReg DO BEGIN
                      economy[i].tax[optper]:= seet[i];
                      economy[i].forestry[optper]:= seef[i]
                  END;
                  FOR i:= 1 TO NoReg DO
                      write(outs,(0.9*economy[i].tax[optper]+0.1*T[i]):6:4,' ');
                  writeln(outs);
                  FOR i:= 1 TO NoReg DO
                      write(outs,(0.9*economy[i].forestry[optper]+0.1*F[i]):6:2,' ');
                  writeln(outs);
              END;
              close(outs);
              stop:= checkstop(cnt)
        UNTIL stop
     END
     ELSE BEGIN
         RunEconomy;
         FOR i:= 1 TO NoReg DO
             economy[i].basewf:= economy[i].npvw;
         CoopCBA(T,F);
         Optimal(T,F)
     END
END;

PROCEDURE RunCoopCEA;

VAR T,F: measure;
    i,psnr: word;
    ceaout: text;

BEGIN
     SetMeasures1;
     assign(ceaout,sessionpathname+'cea.prn');
     rewrite(ceaout);
     IF Sequent THEN BEGIN
        FOR optper:= 0 TO NoPer DO BEGIN
            write(optper:2,' ');
            RunEconomy;
            IF optper < 5 THEN
               psnr:= 4
            ELSE psnr:= NoReg;
            FOR i:= 1 TO NoReg DO
                economy[i].basewf:= economy[i].npvw;
            CoopCEA(psnr,T,F);
            FOR i:= 1 TO NoReg DO
                write(ceaout,T[i]:6:4,' ');
            writeln(ceaout);
            FOR i:= 1 TO NoReg DO
                write(ceaout,F[i]:6:2,' ');
            writeln(ceaout)
        END
     END;
     close(ceaout)
END;

PROCEDURE Main;

VAR i: word;

BEGIN
     Initialise;
     ReInit;
     IF CBA THEN
        IF cooperate THEN
           RunCoopCBA
        ELSE
            RunNonCoopCBA
     ELSE IF CEA THEN
          IF cooperate THEN
             RunCoopCEA
          ELSE
              RunNonCoopCEA
     ELSE BEGIN
          write(' calculating');
          IF Scratch THEN
             SetMeasures1
          ELSE
              SetMeasures4;
          RunEconomy1;
          ShortRep('summary.prn');
          IF cooperate THEN BEGIN
             SetMeasures2;
             FOR i:= 1 TO NoReg DO
                 economy[i].basewf:= economy[i].npvw
          END;
          IF captrans THEN
             Fund
     END;
     Exit
END;

PROCEDURE AltMain;

VAR p,i,r: word;
    taks: double;
    outt, outw: text;

BEGIN
     Initialise;
     assign(outt,sessionpathname+'randtax.prn');
     rewrite(outt);
     assign(outw,sessionpathname+'randwel.prn');
     rewrite(outw);
     FOR p:= 0 TO 20 DO BEGIN
         optper:= p;
         SetMeasures4;
         FOR i:= 1 TO 100 DO BEGIN
             FOR r:= 1 TO NoReg DO BEGIN
                 taks:= random/100;
                 SetTax5(r,p,taks);
                 write(outt,taks:6:4,' ',taks*taks:10:8,' ')
             END;
             reinit;
             runeconomy;
             FOR r:= 1 TO NoReg DO
                 write(outw,economy[r].npvw:8:4,' ');
             writeln(outw);
             writeln(outt)
         END
     END;
     close(outw);
     close(outt)
END;

BEGIN
     Main
END.